
/*
 * author Jiří Zuzaňák
 * institution BUT FIT UPGM
 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#ifdef __WIN32
#include <windows.h>
#else
#include <string.h>
#endif

/*
 * DEBUG_ASSERT - povoli makro debug_assert
 * ASSERT - povoli makro assert
 * MEM_CHECK - povoli kontrolu pridelovani a uvolnovani pameti
 */

#define DEBUG_ASSERT
#define ASSERT
#define MEM_CHECK

/*
 * bit cout of some basic types
 */

const unsigned INT_BIT = sizeof(int)*CHAR_BIT;
const unsigned UINT_BIT = sizeof(unsigned)*CHAR_BIT;

/*
 * c_idx_not_exist - konstanta oznacujici nexistujici index
 * c_array_add - konstanta o kterou se navysuje pamet(nasobena velikosti typu) pridelena dynamickym polim
 */

typedef void * pointer;
const unsigned c_idx_not_exist = UINT_MAX;
const unsigned c_array_add = 64;

/*
 * makro debug_assert slouzi pro debugovani v casto spoustenych usecich kodu ve finalni verzi nebude aktivni
 */
#ifdef DEBUG_ASSERT
#define debug_assert(COND)\
{\
   if (!(COND)) {\
      fprintf(stderr,"DEBUG_ASSERT: %s, line: %d\n",__FILE__,__LINE__);\
      throw;\
   }\
}
#else
#define debug_assert(COND)
#endif

/*
 * makro assert vetsina bude nahrazena chybovou zpravou, pri nesplneni podminky makro vypise 
 * ve kterem souboru a na kterem radku se nachazi, neukoncuje program
 */
#ifdef ASSERT
#define assert(COND)\
{\
   if (!(COND)) {\
      fprintf(stderr,"ASSERT: %s, line: %d\n",__FILE__,__LINE__);\
      throw;\
   }\
}
#else
#define assert(COND)
#endif

/*
 * struktura implementujici kontrolu pameti, na zacatku programu musi byt zavolana funkce mc_init() a
 * na konci mc_clear(), pro alokaci a dealokaci pameti se pouzivaji funkce cmalloc a cfree
 */
#ifdef MEM_CHECK

/* --- definice struktury mce_struct_s ---
 * novy zpusob zpravy pameti pomoci hashovaneho pole
 * udrzuje zaznam o pozici pridelene pameti a jeji velikosti
 * muze byt vypnuta, ale dopad na vykon je minimalni
 */

/*
 * zaznam popisujici blok pameti
 */
struct mce_block_s
{
   void *location;
   unsigned size;
   unsigned next_idx;
};

/*
 * struktura, ktera uchovava seznam zaznamu o pridelene pameti
 */
struct mce_struct_s
{
   unsigned size;
   unsigned data_start;
   unsigned empty_start;
   mce_block_s *data;

   inline void init();
   inline void clear();
   void check();
   void copy_resize(unsigned a_size);
   void *create_block(unsigned a_size);
   void delete_block(void *a_location);
   void print_hash(bool set);
};

/*
 * inicializace kontrolni struktury
 */
inline void mce_struct_s::init()
{
   size = 0;
   data_start = c_idx_not_exist;
   empty_start = c_idx_not_exist;
   data = NULL;
}

/*
 * uvolneni kontrolni struktury z pameti,
 * pred uvolnenim provede kontrolu zda byly vsechny bloky uvolneny
 */
inline void mce_struct_s::clear()
{
   check();

   if (data != NULL) {
      free(data);
   }

   init();
}

/*
 * kontrola zda jsou vsechny bloky pameti uvolneny, pokud ne 
 * zobraz na chybovy vystup informace o neuvolnenych blocich
 */
void mce_struct_s::check()
{
   if (data_start == c_idx_not_exist) {
      return;
   }

   unsigned idx = data_start;
   do {
      mce_block_s &block = data[idx];

      fprintf(stderr,"MEM_CHECK_ENCHANCED: leak at: %lu, size: %d\n",(long unsigned)block.location,block.size);
      free(block.location);

      idx = block.next_idx;
   } while(idx != c_idx_not_exist);
}

/*
 * zmena velikosti pole pro ukladani zaznamu o pridelene pameti
 */
void mce_struct_s::copy_resize(unsigned a_size)
{
   debug_assert(a_size >= size);

   if (a_size == size) {
      return;
   }

   mce_block_s *n_data = (mce_block_s *)malloc(a_size*sizeof(mce_block_s));

   if (size != 0) {
      memcpy(n_data,data,size*sizeof(mce_block_s));
      free(data);
   }

   unsigned idx = size;
   unsigned idx_end = a_size - 1;

   if (idx < idx_end) {
      do {
	 n_data[idx].next_idx = idx + 1;
      } while(++idx < idx_end);
   }

   n_data[idx].next_idx = empty_start;
   empty_start = size;

   data = n_data;
   size = a_size;
}

/*
 * alokuje novy blok pameti, ulozi jej do struktury a
 * vrati ukazatel na tento blok
 */
void *mce_struct_s::create_block(unsigned a_size)
{
   void *location = malloc(a_size);

   if (empty_start == c_idx_not_exist) {
      copy_resize(size + c_array_add);
   }

   unsigned old_empty_start = empty_start;
   mce_block_s &block = data[old_empty_start];

   empty_start = block.next_idx;
   block.next_idx = data_start;
   data_start = old_empty_start;

   block.location = location;
   block.size = a_size;

   return location;
}

/*
 * uvolni alokovany blok z pameti a zaroven zkontroluje jestli
 * byl vubec alokovan
 */
void mce_struct_s::delete_block(void *a_location)
{
   debug_assert(data_start != c_idx_not_exist);

   unsigned idx = data_start;
   unsigned old_idx = c_idx_not_exist;
   do {
      mce_block_s &block = data[idx];

      if (block.location == a_location) {
	 break;
      }

      old_idx = idx;
      idx = block.next_idx;

   } while(idx != c_idx_not_exist);

   debug_assert(idx != c_idx_not_exist);

   mce_block_s &block = data[idx];

   if (old_idx == c_idx_not_exist) {
      data_start = block.next_idx;
   }
   else {
      data[old_idx].next_idx = block.next_idx;
   }

   block.next_idx = empty_start;
   empty_start = idx;

   free(block.location);
}

#define mc_init() mce.init();
#define mc_clear() mce.clear();

#define cmalloc(SIZE) mce.create_block(SIZE)
#define cfree(LOCATION) mce.delete_block(LOCATION)

mce_struct_s mce;

#else

#define mc_init()
#define mc_clear()

#define cmalloc(SIZE) malloc(SIZE)
#define cfree(LOCATION) free(LOCATION)

#endif

